@@ -1,5 +1,5 @@
 /* This file is part of GNU Dico.
-   Copyright (C) 1998-2000, 2008 Sergey Poznyakoff
+   Copyright (C) 1998, 1999, 2000, 2008, 2010 Sergey Poznyakoff
 
    GNU Dico is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -67,7 +67,7 @@ char *help_text;
 dico_list_t /* of struct sockaddr */ listen_addr;
 
 /* User database for AUTH */
-dicod_user_db_t user_db;
+dico_udb_t user_db;
 
 /* Run as this user */
 uid_t user_id;
@@ -77,6 +77,7 @@ dico_list_t /* of gid_t */ group_list;
 
 /* List of directories to search for handler modules. */
 dico_list_t /* of char * */ module_load_path;
+dico_list_t /* of char * */ prepend_load_path;
 
 /* List of configured database module instances */
 dico_list_t /* of dicod_module_instance_t */ modinst_list;
@@ -106,6 +107,9 @@ long ident_timeout = 3; /* Timeout for AUTH I/O */
 /* Provide timing information */
 int timing_option;
 
+/* Name of the default strategy */
+const char *default_strategy_name;
+
 
 /* Configuration */
 
@@ -565,28 +569,6 @@ enable_capability(enum cfg_callback_command cmd,
     return 0;
 }
 
-int
-set_defstrat(enum cfg_callback_command cmd,
-	     dicod_locus_t *locus,
-	     void *varptr,
-	     config_value_t *value,
-	     void *cb_data)
-{
-    if (cmd != callback_set_value) {
-	config_error(locus, 0, _("Unexpected block statement"));
-	return 1;
-    }
-    if (value->type != TYPE_STRING) {
-	config_error(locus, 0, _("expected scalar value but found list"));
-	return 1;
-    }
-    if (dico_set_default_strategy(value->v.string)) {
-	config_error(locus, 0, _("unknown strategy"));
-	return 1;
-    }
-    return 0;
-}
-
 int
 mime_headers_cb (enum cfg_callback_command cmd,
 		 dicod_locus_t *locus,
@@ -657,9 +639,11 @@ struct config_keyword kwd_database[] = {
 
 
 struct user_db_conf {
+    dicod_locus_t locus;
     char *url;
     char *get_pw;
     char *get_groups;
+    char *options;
 };
 
 struct user_db_conf user_db_cfg;
@@ -670,6 +654,9 @@ struct config_keyword kwd_user_db[] = {
     { "group-resource", N_("arg"),
       N_("File containing user group information or a query to retrieve it."),
       cfg_string, NULL, offsetof(struct user_db_conf, get_groups) },
+    { "options", N_("arg"),
+      N_("Implementation-dependent options"),
+      cfg_string, NULL, offsetof(struct user_db_conf, options) },
     { NULL }
 };
 
@@ -685,6 +672,7 @@ user_db_config(enum cfg_callback_command cmd,
     
     switch (cmd) {
     case callback_section_begin:
+	cfg->locus = *locus;
 	if (value->type != TYPE_STRING) 
 	    config_error(locus, 0, _("URL must be a string"));
 	else if (!value->v.string)
@@ -695,19 +683,31 @@ user_db_config(enum cfg_callback_command cmd,
 	break;
 	
     case callback_section_end:
-	udb_create(&user_db, cfg->url, cfg->get_pw, cfg->get_groups, locus);
 	break;
 	
     case callback_set_value:
+	cfg->locus = *locus;
 	if (value->type != TYPE_STRING) 
 	    config_error(locus, 0, _("URL must be a string"));
 	else if (!value->v.string)
 	    config_error(locus, 0, _("empty URL"));
-	udb_create(&user_db, value->v.string, NULL, NULL, locus);
+	cfg->url = strdup(value->v.string);
     }
     return 0;
 }
 
+static void
+init_user_db()
+{
+    if (user_db_cfg.url &&
+	dico_udb_create(&user_db,
+			user_db_cfg.url, user_db_cfg.get_pw,
+			user_db_cfg.get_groups, user_db_cfg.options)) {
+	config_error(&user_db_cfg.locus, errno,
+		     _("cannot create user database"));
+    }
+}
+
 
 int
 alias_cb(enum cfg_callback_command cmd,
@@ -735,7 +735,7 @@ alias_cb(enum cfg_callback_command cmd,
 	    config_error(locus, 0, _("argument %d has wrong type"), i+1);
 	    return 1;
 	}
-	argv[i] = value->v.arg.v[i+1].v.string;
+	argv[i] = (char*) value->v.arg.v[i+1].v.string;
     }
     argv[i] = NULL;
     return alias_install(value->v.arg.v[0].v.string, argc, argv, locus);
@@ -1101,9 +1101,13 @@ struct config_keyword keywords[] = {
     { "module-load-path", N_("path"),
       N_("List of directories searched for database modules."),
       cfg_string|CFG_LIST, &module_load_path },
+    { "prepend-load-path", N_("path"),
+      N_("List of directories searched for database modules prior to "
+	 "the default module directory"),
+      cfg_string|CFG_LIST, &prepend_load_path },
     { "default-strategy", N_("name"),
       N_("Set the name of the default matching strategy."),
-      cfg_string, NULL, 0, set_defstrat },
+      cfg_string, &default_strategy_name },
     { "timing", N_("arg"),
       N_("Provide timing information after successful completion of an "
 	 "operation."),
@@ -1351,6 +1355,14 @@ apply_conf_override(struct dicod_conf_override *ovr)
 	transcript = ovr->transcript;
 }
 
+
+static void
+udb_init()
+{
+    dico_udb_define(&text_udb_def);
+}
+
+
 
 int
 main(int argc, char **argv)
@@ -1383,20 +1395,20 @@ main(int argc, char **argv)
     get_options(argc, argv, &ovr);
 
     if (mode == MODE_PREPROC)
-	return preprocess_config(preprocessor);
+	exit(preprocess_config(preprocessor) ? EX_UNAVAILABLE : 0);
 
     config_set_keywords(keywords);
     if (config_parse(config_file))
-	exit(1);
+	exit(EX_CONFIG);
 
     apply_conf_override(&ovr);
     
     register_sasl();
     if (dicod_capa_flush())
-	exit(1);
+	exit(EX_SOFTWARE);
     compile_access_log();
     if (config_lint_option)
-	exit(0);
+	exit(EX_OK);
     
     dicod_log_setup();
     
@@ -1404,15 +1416,23 @@ main(int argc, char **argv)
     
     begin_timing("server");
     dicod_server_init();
+    init_user_db();
     flush_strat_forward();
+    if (default_strategy_name
+	&& dico_set_default_strategy(default_strategy_name)) {
+	dico_die(EX_UNAVAILABLE, L_ERR, 0, _("unknown strategy"));
+    }
+	
     markup_flush_capa();
+
     switch (mode) {
     case MODE_DAEMON:
 	dicod_server(argc, argv);
 	break;
 	
     case MODE_INETD:
-	rc = dicod_inetd();
+	if (dicod_inetd())
+	    rc = EX_UNAVAILABLE;
 	dicod_server_cleanup();
 	break;
     }
