@@ -1,5 +1,5 @@
 /* This file is part of GNU Dico.
-   Copyright (C) 1998-2000, 2008 Sergey Poznyakoff
+   Copyright (C) 1998, 1999, 2000, 2008, 2010 Sergey Poznyakoff
 
    GNU Dico is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -18,9 +18,10 @@
 #endif
 
 #include <xdico.h>
+#include <sysexits.h>
 
 void
 xalloc_die()
 {
-    dico_die(1, L_CRIT, 0, "Not enough memory");
+    dico_die(EX_OSERR, L_CRIT, 0, "Not enough memory");
 }
