@@ -1,5 +1,5 @@
 /* This file is part of GNU Dico.
-   Copyright (C) 2008, 2009 Sergey Poznyakoff
+   Copyright (C) 2008, 2009, 2010 Sergey Poznyakoff
 
    GNU Dico is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -140,11 +140,11 @@ parse_index_entry(const char *filename, size_t line,
     
     memset(&itr, 0, sizeof(itr));
     
-    utf8_iter_first(&itr, (unsigned char *)buf);
+    utf8_iter_first(&itr, buf);
 
     rc = 0;
     for (nfield = 0; nfield < 3; nfield++) {
-	unsigned char *start, *end;
+	char *start, *end;
 	size_t len;
 	
 	/* Skip whitespace */
@@ -232,10 +232,9 @@ static int
 read_index(struct dictdb *db, const char *idxname, int tws)
 {
     struct stat st;
-    FILE *fp;
-    char buf[512]; /* FIXME: fixed size */
     int rc;
     dico_list_t list;
+    dico_stream_t stream;
     
     if (stat(idxname, &st)) {
 	dico_log(L_ERR, errno, _("open_index: cannot stat `%s'"), idxname);
@@ -246,11 +245,21 @@ read_index(struct dictdb *db, const char *idxname, int tws)
 		 idxname);
 	return 1;
     }
-    fp = fopen(idxname, "r");
-    if (!fp) {
-	dico_log(L_ERR, errno, _("open_index: cannot open `%s'"), idxname);
+
+
+    stream = dico_mapfile_stream_create(idxname, DICO_STREAM_READ);
+    if (!stream) {
+	dico_log(L_ERR, errno,
+		 _("cannot create stream `%s'"), idxname);
 	return 1;
     }
+    rc = dico_stream_open(stream);
+    if (rc) {
+	dico_log(L_ERR, 0,
+		 _("cannot open stream `%s': %s"),
+		 idxname, dico_stream_strerror(stream, rc));
+	dico_stream_destroy(&stream);
+    }
 
     list = dico_list_create();
     if (!list) {
@@ -260,17 +269,21 @@ read_index(struct dictdb *db, const char *idxname, int tws)
 	dico_iterator_t itr;
 	size_t i;
 	struct index_entry *ep;
+	char *buf = NULL;
+	size_t bufsize = 0;
+	size_t rdsize;
 
 	rc = 0;
 
 	i = 0;
-	while (fgets(buf, sizeof(buf), fp)) {
+	while (!dico_stream_getline(stream, &buf, &bufsize, &rdsize)) {
 	    i++;
 	    dico_trim_nl(buf);
 	    rc = parse_index_entry(idxname, i, list, buf, tws);
 	    if (rc)
 		break;
 	}
+	free(buf);
 	if (rc) {
 	    dico_list_set_free_item(list, free_index_entry, NULL);
 	} else {
@@ -288,7 +301,8 @@ read_index(struct dictdb *db, const char *idxname, int tws)
 	dico_list_destroy(&list);
     }
 
-    fclose(fp);
+    dico_stream_close(stream);
+    dico_stream_destroy(&stream);
     return rc;
 }
 
@@ -439,7 +453,7 @@ revert_word(char *dst, const char *src, size_t len)
     char *p = dst + len;
 
     *p = 0;
-    for (utf8_iter_first(&itr, (unsigned char *)src);
+    for (utf8_iter_first(&itr, (char *)src);
 	 !utf8_iter_end_p(&itr);
 	 utf8_iter_next(&itr)) {
 	p -= itr.curwidth;
@@ -521,9 +535,27 @@ register_strategies(void)
     return 0;
 }
 
+static int
+compare_entry(const void *a, const void *b)
+{
+    const struct index_entry *epa = a;
+    const struct index_entry *epb = b;
+    compare_count++;
+    return utf8_strcasecmp(epa->word, epb->word);
+}
+
+static int
+compare_entry_ptr(const void *a, const void *b)
+{
+    const struct index_entry *epa = *(const struct index_entry **)a;
+    const struct index_entry *epb = *(const struct index_entry **)b;
+    return compare_entry(epa, epb);
+}
+
 static int
 common_match(struct dictdb *db, const char *word,
-	     int (*compare) (const void *, const void *), struct result *res)
+	     int (*compare) (const void *, const void *),
+	     int unique, struct result *res)
 {
     struct index_entry x, *ep;
     
@@ -547,6 +579,13 @@ common_match(struct dictdb *db, const char *word,
 	    memerr("common_match");
 	    return 0;
 	}
+	res->itr = NULL;
+	if (unique) {
+	    dico_list_set_comparator(res->list,
+				     (int (*)(const void *, void *))
+				          compare_entry);
+	    dico_list_set_flags(res->list, DICO_LIST_COMPARE_TAIL);
+	}
 	for (p++; p < ep; p++) 
 	    if (!RESERVED_WORD(db, p->word))
 		dico_list_append(res->list, p);
@@ -557,27 +596,10 @@ common_match(struct dictdb *db, const char *word,
 }
 
 
-static int
-compare_entry(const void *a, const void *b)
-{
-    const struct index_entry *epa = a;
-    const struct index_entry *epb = b;
-    compare_count++;
-    return utf8_strcasecmp(epa->word, epb->word);
-}
-
-static int
-compare_entry_ptr(const void *a, const void *b)
-{
-    const struct index_entry *epa = *(const struct index_entry **)a;
-    const struct index_entry *epb = *(const struct index_entry **)b;
-    return compare_entry(epa, epb);
-}
-
 static int
 exact_match(struct dictdb *db, const char *word, struct result *res)
 {
-    return common_match(db, word, compare_entry, res);
+    return common_match(db, word, compare_entry, 1, res);
 }
 
 static int
@@ -595,7 +617,7 @@ compare_prefix(const void *a, const void *b)
 static int
 prefix_match(struct dictdb *db, const char *word, struct result *res)
 {
-    return common_match(db, word, compare_prefix, res);
+    return common_match(db, word, compare_prefix, 1, res);
 }
 
 static int
@@ -674,14 +696,17 @@ suffix_match(struct dictdb *db, const char *word, struct result *res)
 	    free(tmp);
 	    return 1;
 	}
-
+	dico_list_set_comparator(list,
+				 (int (*)(const void *, void *))
+				   compare_entry);
+	dico_list_set_flags(list, DICO_LIST_COMPARE_TAIL);
 	for (i = 0; i < count; i++) 
-	    if (i == 0 || compare_entry(tmp[i-1], tmp[i])) 
-		dico_list_append(list, tmp[i]);
+	    dico_list_append(list, tmp[i]);
      
 	free(tmp);
 	res->type = result_match;
 	res->list = list;
+	res->itr = NULL;
 	res->compare_count = compare_count;
 	rc = 0;
     } else 
@@ -775,12 +800,12 @@ _match_simple(struct dictdb *db, const char *strat, const char *word)
 }
 
 static dico_result_t
-_match_all(struct dictdb *db, const char *word,
-	   dico_select_t sel, void *closure)
+_match_all(struct dictdb *db, dico_strategy_t strat, const char *word)
 {
     dico_list_t list;
     size_t count, i;
     struct result *res;
+    struct dico_key key;
     
     list = dico_list_create();
 
@@ -789,18 +814,24 @@ _match_all(struct dictdb *db, const char *word,
 	return NULL;
     }
 
-    if (sel(DICO_SELECT_BEGIN, word, NULL, closure)) {
-	dico_log(L_ERR, 0, _("_match_all: initial select failed"));
+    dico_list_set_comparator(list,
+			     (int (*)(const void *, void *))
+			       compare_entry);
+    dico_list_set_flags(list, DICO_LIST_COMPARE_TAIL);
+
+    if (dico_key_init(&key, strat, word)) {
+	dico_log(L_ERR, 0, _("_match_all: key initialization failed"));
 	return NULL;
     }
     
     for (i = 0; i < db->numwords; i++) 
 	if (!RESERVED_WORD(db, db->index[i].word)
-	    && sel(DICO_SELECT_RUN, word, db->index[i].word, closure)) 
+	    && dico_key_match(&key, db->index[i].word)) 
 	    dico_list_append(list, &db->index[i]);
-	
+
+    dico_key_deinit(&key);
+    
     compare_count = db->numwords;
-    sel(DICO_SELECT_END, word, NULL, closure);
 	
     count = dico_list_count(list);
     if (count == 0) {
@@ -814,6 +845,7 @@ _match_all(struct dictdb *db, const char *word,
     res->db = db;
     res->type = result_match;
     res->list = list;
+    res->itr = NULL;
     res->compare_count = compare_count;
     return (dico_result_t) res;
 }
@@ -827,7 +859,7 @@ mod_match(dico_handle_t hp, const dico_strategy_t strat, const char *word)
 	return NULL;
     
     if (strat->sel) 
-	return _match_all(db, word, strat->sel, strat->closure);
+	return _match_all(db, strat, word);
     else
 	return _match_simple(db, strat->name, word);
 }
@@ -841,7 +873,7 @@ mod_define(dico_handle_t hp, const char *word)
 
     if (RESERVED_WORD(db, word))
 	return NULL;
-    rc = exact_match(db, word, &res);
+    rc = common_match(db, word, compare_entry, 0, &res);
     if (rc)
 	return NULL;
     rp = malloc(sizeof(*rp));
@@ -886,7 +918,15 @@ static int
 mod_output_result (dico_result_t rp, size_t n, dico_stream_t str)
 {
     struct result *res = (struct result *) rp;
-    const struct index_entry *ep = dico_list_item(res->list, n);
+    const struct index_entry *ep;
+
+    if (!res->itr) {
+	res->itr = dico_list_iterator(res->list);
+	if (!res->itr)
+	    return 1;
+    }
+    ep = dico_iterator_item(res->itr, n);
+    
     
     switch (res->type) {
     case result_match:
@@ -917,6 +957,7 @@ static void
 mod_free_result(dico_result_t rp)
 {
     struct result *res = (struct result *) rp;
+    dico_iterator_destroy(&res->itr);
     dico_list_destroy(&res->list);
     free(rp);
 }
