@@ -1,5 +1,5 @@
 /* This file is part of GNU Dico.
-   Copyright (C) 2008 Sergey Poznyakoff
+   Copyright (C) 2008, 2010 Sergey Poznyakoff
 
    GNU Dico is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -26,13 +26,23 @@
 #include <libguile.h>
 #include <appi18n.h>
 
+#ifndef HAVE_SCM_T_OFF
+typedef off_t scm_t_off;
+#endif
+
 
 /* General-purpose eval handlers */
 
-static SCM
-eval_catch_body(void *list)
+struct apply_data {
+	SCM proc;
+	SCM arg;
+};
+
+SCM
+apply_catch_body(void *data)
 {
-    return scm_primitive_eval((SCM)list);
+    struct apply_data *xp = data;
+    return scm_apply_0(xp->proc, xp->arg);
 }
 
 static SCM
@@ -61,7 +71,7 @@ guile_safe_exec(SCM (*handler) (void *data), void *data, SCM *result)
 {
     jmp_buf jmp_env;
     struct scheme_exec_data ed;
-	
+
     if (setjmp(jmp_env))
 	return 1;
     ed.handler = handler;
@@ -84,9 +94,9 @@ static SCM
 load_path_handler(void *data)
 {
     struct load_closure *lp = data;
-    
+
     scm_set_program_arguments(lp->argc, lp->argv, lp->filename);
-    scm_primitive_load_path(scm_makfrom0str(lp->filename));
+    scm_primitive_load_path(scm_from_locale_string(lp->filename));
     return SCM_UNDEFINED;
 }
 
@@ -113,7 +123,7 @@ _add_load_path(char *path)
 {
     SCM scm, path_scm;
     SCM *pscm;
-    
+
     path_scm = SCM_VARIABLE_REF(scm_c_lookup("%load-path"));
     for (scm = path_scm; scm != SCM_EOL; scm = SCM_CDR(scm)) {
 	SCM val = SCM_CAR(scm);
@@ -128,7 +138,7 @@ _add_load_path(char *path)
 
     pscm = SCM_VARIABLE_LOC(scm_c_lookup("%load-path"));
     *pscm = scm_append(scm_list_3(path_scm,
-				  scm_list_1(scm_makfrom0str(path)),
+				  scm_list_1(scm_from_locale_string(path)),
 				  SCM_EOL));
 }
 
@@ -142,7 +152,7 @@ static char *
 proc_name(SCM proc)
 {
     return scm_to_locale_string(
-	          scm_symbol_to_string(scm_procedure_name(proc)));
+		  scm_symbol_to_string(scm_procedure_name(proc)));
 }
 
 static void
@@ -163,8 +173,8 @@ static int
 guile_call_proc(SCM *result, SCM proc, SCM arglist)
 {
     jmp_buf jmp_env;
-    SCM cell;
-    
+    struct apply_data adata;
+
     if (setjmp(jmp_env)) {
 	char *name = proc_name(proc);
 	dico_log(L_NOTICE, 0,
@@ -173,15 +183,85 @@ guile_call_proc(SCM *result, SCM proc, SCM arglist)
 	free(name);
 	return 1;
     }
-    cell = scm_cons(proc, arglist);
+    adata.proc = proc;
+    adata.arg = arglist;
     *result = scm_internal_lazy_catch(SCM_BOOL_T,
-				      eval_catch_body, cell,
+				      apply_catch_body, &adata,
 				      eval_catch_handler, &jmp_env);
     return 0;
 }
 
 
-long _guile_strategy_tag;
+scm_t_bits _guile_dico_key_tag;
+
+static SCM
+dico_new_scm_key(struct dico_key **pkey)
+{
+    struct dico_key *kptr;
+
+    kptr = scm_gc_malloc (sizeof (*kptr), "Dico key");
+    *pkey = kptr;
+    SCM_RETURN_NEWSMOB(_guile_dico_key_tag, kptr);
+}
+
+static scm_sizet
+_guile_dico_key_free(SCM message_smob)
+{
+    struct dico_key *kp = (struct dico_key *) SCM_CDR (message_smob);
+    dico_key_deinit(kp);
+    free(kp);
+    return 0;
+}
+
+static int
+_guile_dico_key_print(SCM message_smob, SCM port, scm_print_state *pstate)
+{
+    struct dico_key *kp = (struct dico_key *) SCM_CDR (message_smob);
+    scm_puts("#<key ", port);
+    scm_puts(kp->strat->name, port);
+    scm_puts(" (", port);
+    scm_puts(kp->word, port);
+    scm_puts(")>", port);
+    return 1;
+}
+
+static void
+_guile_init_dico_key()
+{
+    _guile_dico_key_tag =
+	scm_make_smob_type("Dico key", sizeof (struct dico_key));
+    scm_set_smob_free(_guile_dico_key_tag, _guile_dico_key_free);
+    scm_set_smob_print(_guile_dico_key_tag, _guile_dico_key_print);
+}
+
+#define CELL_IS_KEY(s) \
+    (!SCM_IMP(s) && SCM_CELL_TYPE(s) == _guile_dico_key_tag)
+
+SCM_DEFINE_PUBLIC(scm_dico_key_p, "dico-key?",
+		  1, 0, 0,
+		  (SCM obj),
+		  "Return @samp{#t} if @var{obj} is a selection key.")
+#define FUNC_NAME s_scm_dico_key_p
+{
+    return CELL_IS_KEY(obj) ? SCM_BOOL_T : SCM_BOOL_F;
+}
+#undef FUNC_NAME
+
+SCM_DEFINE_PUBLIC(scm_dico_key__word, "dico-key->word",
+		  1, 0, 0,
+		  (SCM key),
+		  "Return search word from the @var{key}.")
+#define FUNC_NAME s_scm_dico_key__word
+{
+    struct dico_key *kp;
+    SCM_ASSERT(CELL_IS_KEY(key), key, SCM_ARG1, FUNC_NAME);
+    kp = (struct dico_key *) SCM_CDR(key);
+    return scm_from_locale_string(kp->word);
+}
+#undef FUNC_NAME
+
+
+scm_t_bits _guile_strategy_tag;
 
 struct _guile_strategy
 {
@@ -201,10 +281,10 @@ _make_strategy(const dico_strategy_t strat)
 static scm_sizet
 _guile_strategy_free(SCM message_smob)
 {
-  struct _guile_strategy *sp =
-      (struct _guile_strategy *) SCM_CDR (message_smob);
-  free(sp);
-  return 0;
+    struct _guile_strategy *sp =
+	(struct _guile_strategy *) SCM_CDR (message_smob);
+    free(sp);
+    return 0;
 }
 
 static int
@@ -232,140 +312,184 @@ _guile_init_strategy()
 #define CELL_IS_STRAT(s) \
     (!SCM_IMP(s) && SCM_CELL_TYPE(s) == _guile_strategy_tag)
 
-SCM_DEFINE(scm_dico_strat_selector_p, "dico-strat-selector?", 1, 0, 0,
-	   (SCM STRAT),
-	   "Return true if STRAT has a selector.")
+SCM_DEFINE_PUBLIC(scm_dico_strat_selector_p, "dico-strat-selector?", 1, 0, 0,
+		  (SCM strat),
+		  "Return true if @var{strat} has a selector.")
 #define FUNC_NAME s_scm_dico_strat_selector_p
 {
     struct _guile_strategy *sp;
-    
-    SCM_ASSERT(CELL_IS_STRAT(STRAT), STRAT, SCM_ARG1, FUNC_NAME);
-    sp = (struct _guile_strategy *) SCM_CDR(STRAT);
+
+    SCM_ASSERT(CELL_IS_STRAT(strat), strat, SCM_ARG1, FUNC_NAME);
+    sp = (struct _guile_strategy *) SCM_CDR(strat);
     return sp->strat->sel ? SCM_BOOL_T : SCM_BOOL_F;
 }
 #undef FUNC_NAME
 
-SCM_DEFINE(scm_dico_strat_select_p, "dico-strat-select?", 3, 0, 0,
-	   (SCM STRAT, SCM WORD, SCM KEY),
-	   "Return true if KEY matches WORD as per strategy selector STRAT.")
+SCM_DEFINE_PUBLIC(scm_dico_strat_select_p, "dico-strat-select?", 3, 0, 0,
+		  (SCM strat, SCM word, SCM key),
+		  "Return true if @var{key} matches @var{word} as per strategy selector @var{strat}.")
 #define FUNC_NAME s_scm_dico_strat_select_p
 {
     struct _guile_strategy *sp;
-    char *key, *word;
+    struct dico_strategy *stratp;
+    char *wordstr;
     int rc;
+
+    SCM_ASSERT(CELL_IS_STRAT(strat), strat, SCM_ARG1, FUNC_NAME);
+    SCM_ASSERT(scm_is_string(word), word, SCM_ARG2, FUNC_NAME);
+
+    sp = (struct _guile_strategy *) SCM_CDR(strat);
+    stratp = sp->strat;
     
-    SCM_ASSERT(CELL_IS_STRAT(STRAT), STRAT, SCM_ARG1, FUNC_NAME);
-    SCM_ASSERT(scm_is_string(WORD), WORD, SCM_ARG2, FUNC_NAME);
-    SCM_ASSERT(scm_is_string(KEY), KEY, SCM_ARG3, FUNC_NAME);
-    sp = (struct _guile_strategy *) SCM_CDR(STRAT);
-    key = scm_to_locale_string(KEY);
-    word = scm_to_locale_string(WORD);
-    rc = sp->strat->sel(DICO_SELECT_RUN, key, word, sp->strat->closure);
-    free(key);
-    free(word);
+    if (scm_is_string(key)) {
+	char *keystr = scm_to_locale_string(key);
+	struct dico_key skey;
+
+	if (dico_key_init(&skey, stratp, wordstr)) {
+	    free(keystr);
+	    scm_misc_error(FUNC_NAME,
+			   "key initialization failed: ~S",
+			   scm_list_1(key));
+	}
+	rc = dico_key_match(&skey, wordstr);
+	dico_key_deinit(&skey);
+	free(wordstr);
+	free(keystr);
+    } else {
+	struct dico_key *kptr;
+
+	SCM_ASSERT(CELL_IS_KEY(key), key, SCM_ARG3, FUNC_NAME);
+    
+	kptr = (struct dico_key *) SCM_CDR(key);
+	wordstr = scm_to_locale_string(word);
+	rc = dico_key_match(kptr, wordstr);
+	free(wordstr);
+    }
     return rc ? SCM_BOOL_T : SCM_BOOL_F;
 }
 #undef FUNC_NAME
 
-SCM_DEFINE(scm_dico_strat_name, "dico-strat-name", 1, 0, 0,
-	   (SCM STRAT),
-	   "Return the name of the strategy STRAT.")
+SCM_DEFINE_PUBLIC(scm_dico_strat_name, "dico-strat-name", 1, 0, 0,
+		  (SCM strat),
+		  "Return the name of the strategy @var{strat}.")
 #define FUNC_NAME s_scm_dico_strat_name
 {
     struct _guile_strategy *sp;
-    
-    SCM_ASSERT(CELL_IS_STRAT(STRAT), STRAT, SCM_ARG1, FUNC_NAME);
-    sp = (struct _guile_strategy *) SCM_CDR(STRAT);
-    return scm_makfrom0str(sp->strat->name);
+
+    SCM_ASSERT(CELL_IS_STRAT(strat), strat, SCM_ARG1, FUNC_NAME);
+    sp = (struct _guile_strategy *) SCM_CDR(strat);
+    return scm_from_locale_string(sp->strat->name);
 }
 #undef FUNC_NAME
 
-SCM_DEFINE(scm_dico_strat_description, "dico-strat-description", 1, 0, 0,
-	   (SCM STRAT),
-	   "Return a textual description of the strategy STRAT.")
+SCM_DEFINE_PUBLIC(scm_dico_strat_description, "dico-strat-description",
+		  1, 0, 0,
+		  (SCM strat),
+		  "Return a textual description of the strategy @var{strat}.")
 #define FUNC_NAME s_scm_dico_strat_description
 {
     struct _guile_strategy *sp;
-    
-    SCM_ASSERT(CELL_IS_STRAT(STRAT), STRAT, SCM_ARG1, FUNC_NAME);
-    sp = (struct _guile_strategy *) SCM_CDR(STRAT);
-    return scm_makfrom0str(sp->strat->descr);
+
+    SCM_ASSERT(CELL_IS_STRAT(strat), strat, SCM_ARG1, FUNC_NAME);
+    sp = (struct _guile_strategy *) SCM_CDR(strat);
+    return scm_from_locale_string(sp->strat->descr);
 }
 #undef FUNC_NAME
 
-SCM_DEFINE(scm_dico_strat_default_p, "dico-strat-default?", 1, 0, 0,
-	   (SCM STRAT),
-	   "Return true if STRAT is a default strategy.")
+SCM_DEFINE_PUBLIC(scm_dico_strat_default_p, "dico-strat-default?", 1, 0, 0,
+		  (SCM strat),
+		  "Return true if @var{strat} is a default strategy.")
 #define FUNC_NAME s_scm_dico_strat_default_p
 {
     struct _guile_strategy *sp;
-    
-    SCM_ASSERT(CELL_IS_STRAT(STRAT), STRAT, SCM_ARG1, FUNC_NAME);
-    sp = (struct _guile_strategy *) SCM_CDR(STRAT);
+
+    SCM_ASSERT(CELL_IS_STRAT(strat), strat, SCM_ARG1, FUNC_NAME);
+    sp = (struct _guile_strategy *) SCM_CDR(strat);
     return dico_strategy_is_default_p(sp->strat) ? SCM_BOOL_T : SCM_BOOL_F;
 }
 #undef FUNC_NAME
 
 
-static SCM
-_scm_guile_sel(void *data)
+SCM_DEFINE_PUBLIC(scm_dico_make_key, "dico-make-key",
+		  2, 0, 0,
+		  (SCM strat, SCM word),
+		  "Make a key for given @var{word} and strategy @var{strat}.")
+#define FUNC_NAME s_scm_dico_make_key
 {
-    return scm_primitive_eval((SCM)data);
+    SCM ret;
+    struct dico_key *key;
+    struct _guile_strategy *sp;
+    char *wordstr;
+    int rc;
+    
+    SCM_ASSERT(CELL_IS_STRAT(strat), strat, SCM_ARG1, FUNC_NAME);
+    SCM_ASSERT(scm_is_string(word), word, SCM_ARG2, FUNC_NAME);
+    sp = (struct _guile_strategy *) SCM_CDR(strat);
+    wordstr = scm_to_locale_string(word);
+    ret = dico_new_scm_key(&key);
+    rc = dico_key_init(key, sp->strat, wordstr);
+    free(wordstr);
+    if (rc)
+	scm_misc_error(FUNC_NAME,
+		       "key initialization failed: ~S",
+		       scm_list_1(ret));
+    return ret;
 }
+#undef FUNC_NAME
 
+
 static int
-_guile_selector(int cmd, const char *word, const char *dict_word,
-		void *closure)
+_guile_selector(int cmd, struct dico_key *key, const char *dict_word)
 {
     SCM result;
-    SCM list = scm_list_4((SCM)closure,
-			  scm_cons(SCM_IM_QUOTE, scm_from_int (cmd)),
-			  scm_cons(SCM_IM_QUOTE, scm_makfrom0str(word)),
-			  scm_cons(SCM_IM_QUOTE, scm_makfrom0str(dict_word)));
-    if (guile_safe_exec(_scm_guile_sel, list, &result))
+    SCM list = scm_list_4((SCM)key->strat->closure,
+			  scm_from_int (cmd),
+			  scm_from_locale_string(key->word),
+			  scm_from_locale_string(dict_word));
+    if (guile_safe_exec(apply_catch_body, list, &result))
 	return 0;
     return result != SCM_BOOL_F;
 }
 
-SCM_DEFINE(scm_dico_register_strat, "dico-register-strat", 2, 1, 0,
-	   (SCM STRAT, SCM DESCR, SCM FUN),
-	   "Register a new strategy.")
+SCM_DEFINE_PUBLIC(scm_dico_register_strat, "dico-register-strat", 2, 1, 0,
+		  (SCM strat, SCM descr, SCM fun),
+		  "Register a new strategy.")
 #define FUNC_NAME s_scm_dico_register_strat
 {
-    struct dico_strategy strat;
-	
-    SCM_ASSERT(scm_is_string(STRAT), STRAT, SCM_ARG1, FUNC_NAME);
-    SCM_ASSERT(scm_is_string(DESCR), DESCR, SCM_ARG2, FUNC_NAME);
+    struct dico_strategy strategy;
 
-    if (!SCM_UNBNDP(FUN))
-	SCM_ASSERT(scm_procedure_p(FUN), FUN, SCM_ARG3, FUNC_NAME);
+    SCM_ASSERT(scm_is_string(strat), strat, SCM_ARG1, FUNC_NAME);
+    SCM_ASSERT(scm_is_string(descr), descr, SCM_ARG2, FUNC_NAME);
 
-    strat.name = scm_to_locale_string(STRAT);
-    strat.descr = scm_to_locale_string(DESCR);
-    if (SCM_UNBNDP(FUN)) {
-	strat.sel = NULL;
-	strat.closure = NULL;
+    if (!SCM_UNBNDP(fun))
+	SCM_ASSERT(scm_procedure_p(fun), fun, SCM_ARG3, FUNC_NAME);
+
+    strategy.name = scm_to_locale_string(strat);
+    strategy.descr = scm_to_locale_string(descr);
+    if (SCM_UNBNDP(fun)) {
+	strategy.sel = NULL;
+	strategy.closure = NULL;
     } else {
-	strat.sel = _guile_selector;
-	strat.closure = FUN;
+	strategy.sel = _guile_selector;
+	strategy.closure = fun;
     }
-    dico_strategy_add(&strat);
-    free(strat.name);
-    free(strat.descr);
+    dico_strategy_add(&strategy);
+    free(strategy.name);
+    free(strategy.descr);
     return SCM_UNSPECIFIED;
 }
 #undef FUNC_NAME
 
 
-SCM_DEFINE(scm_dico_register_markup, "dico-register-markup", 1, 0, 0,
-	   (SCM TYPE),
-	   "Register new markup type.")
+SCM_DEFINE_PUBLIC(scm_dico_register_markup, "dico-register-markup", 1, 0, 0,
+		  (SCM type),
+		  "Register new markup type.")
 #define FUNC_NAME s_scm_dico_register_markup
 {
     int rc;
     char *str;
-    SCM_ASSERT(scm_is_string(TYPE), TYPE, SCM_ARG1, FUNC_NAME);
-    str = scm_to_locale_string(TYPE);
+    SCM_ASSERT(scm_is_string(type), type, SCM_ARG1, FUNC_NAME);
+    str = scm_to_locale_string(type);
     rc = dico_markup_register(str);
     free(str);
     switch (rc) {
@@ -378,7 +502,7 @@ SCM_DEFINE(scm_dico_register_markup, "dico-register-markup", 1, 0, 0,
     case EINVAL:
 	scm_misc_error(FUNC_NAME,
 		       "Invalid markup name: ~S",
-		       scm_list_1(TYPE));
+		       scm_list_1(type));
 
     default:
 	scm_misc_error(FUNC_NAME,
@@ -389,17 +513,17 @@ SCM_DEFINE(scm_dico_register_markup, "dico-register-markup", 1, 0, 0,
 }
 #undef FUNC_NAME
 
-SCM_DEFINE(scm_dico_current_markup, "dico-current-markup", 0, 0, 0,
-	   (),
-	   "Return current dico markup type.")
+SCM_DEFINE_PUBLIC(scm_dico_current_markup, "dico-current-markup", 0, 0, 0,
+		  (),
+		  "Return current dico markup type.")
 #define FUNC_NAME s_scm_dico_current_markup
 {
-    return scm_makfrom0str(dico_markup_type);
+    return scm_from_locale_string(dico_markup_type);
 }
 #undef FUNC_NAME
-    
+
 
-static long scm_tc16_dico_port;
+static scm_t_bits scm_tc16_dico_port;
 struct _guile_dico_port {
     dico_stream_t str;
 };
@@ -413,9 +537,9 @@ _make_dico_port(dico_stream_t str)
 
     dp = scm_gc_malloc (sizeof (struct _guile_dico_port), "dico-port");
     dp->str = str;
-    port = scm_cell(scm_tc16_dico_port, 0);
-    pt = scm_add_to_port_table(port);
-    SCM_SETPTAB_ENTRY(port, pt);
+
+    port = scm_new_port_table_entry(scm_tc16_dico_port);
+    pt = SCM_PTAB_ENTRY(port);
     pt->rw_random = 0;
     SCM_SET_CELL_TYPE(port,
 		      (scm_tc16_dico_port | SCM_OPN | SCM_WRTNG | SCM_BUF0));
@@ -472,11 +596,11 @@ _dico_port_write(SCM port, const void *data, size_t size)
     dico_stream_write(dp->str, data, size);
 }
 
-static off_t
-_dico_port_seek (SCM port, off_t offset, int whence)
+static scm_t_off
+_dico_port_seek (SCM port, scm_t_off offset, int whence)
 {
     struct _guile_dico_port *dp = DICO_PORT(port);
-    return dico_stream_seek(dp->str, offset, whence);
+    return (scm_t_off) dico_stream_seek(dp->str, (off_t) offset, whence);
 }
 
 static int
@@ -498,11 +622,11 @@ _guile_init_dico_port()
     scm_set_port_flush (scm_tc16_dico_port, _dico_port_flush);
     scm_set_port_close (scm_tc16_dico_port, _dico_port_close);
     scm_set_port_seek (scm_tc16_dico_port, _dico_port_seek);
-}    
+}
 
 
 
-static long scm_tc16_dico_log_port;
+static scm_t_bits scm_tc16_dico_log_port;
 
 static SCM
 _make_dico_log_port(int level)
@@ -530,37 +654,13 @@ _guile_init_dico_log_port()
     scm_set_port_flush (scm_tc16_dico_log_port, _dico_port_flush);
     scm_set_port_close (scm_tc16_dico_log_port, _dico_port_close);
     scm_set_port_seek (scm_tc16_dico_log_port, _dico_port_seek);
-}    
+}
 
 
 static void
 _guile_init_funcs (void)
 {
-    scm_c_define_gsubr(s_scm_dico_strat_selector_p, 1, 0, 0,
-		       scm_dico_strat_selector_p);
-    scm_c_define_gsubr(s_scm_dico_strat_select_p, 3, 0, 0,
-		       scm_dico_strat_select_p);
-    scm_c_define_gsubr(s_scm_dico_strat_name, 1, 0, 0,
-		       scm_dico_strat_name);
-    scm_c_define_gsubr(s_scm_dico_strat_description, 1, 0, 0,
-		       scm_dico_strat_description);
-    scm_c_define_gsubr(s_scm_dico_strat_default_p, 1, 0, 0,
-		       scm_dico_strat_default_p);
-    scm_c_define_gsubr(s_scm_dico_register_strat, 2, 1, 0,
-		       scm_dico_register_strat);
-    scm_c_define_gsubr(s_scm_dico_register_markup, 1, 0, 0,
-		       scm_dico_register_markup);
-    scm_c_define_gsubr(s_scm_dico_current_markup, 0, 0, 0,
-		       scm_dico_current_markup);
-    scm_c_export("dico-strat-selector?", 
-		 "dico-strat-select?",
-		 "dico-strat-name",
-		 "dico-strat-description",
-	         "dico-strat-default?",
-		 "dico-register-strat",
-		 "dico-register-markup",
-		 "dico-current-markup",
-		 NULL);
+#include <guile.x>
 }
 
 
@@ -636,11 +736,10 @@ call_init_handler(void *data)
     SCM procsym = SCM_VARIABLE_REF(scm_c_lookup(p->init_fun));
     SCM arg;
     if (p->db_name)
-	arg = scm_makfrom0str(p->db_name);
+	arg = scm_from_locale_string(p->db_name);
     else
 	arg = SCM_BOOL_F;
-    return scm_primitive_eval(scm_list_2(procsym,
-					 scm_cons(SCM_IM_QUOTE, arg)));
+    return scm_apply_0(procsym, scm_list_1(arg));
 }
 
 static int
@@ -653,7 +752,7 @@ init_vtab(const char *init_fun, const char *dbname, guile_vtab vtab)
     istr.db_name = dbname;
     if (guile_safe_exec(call_init_handler, &istr, &res))
 	return 1;
-    
+
     if (!scm_is_pair(res) && res != SCM_EOL) {
 	str_rettype_error(init_fun);
 	return 1;
@@ -690,7 +789,7 @@ set_load_path(struct dico_option *opt, const char *val)
     char *tmp = strdup(val);
     if (!tmp)
 	return 1;
-    for (p = strtok(tmp, ":"); p; p = strtok(NULL, ":")) 
+    for (p = strtok(tmp, ":"); p; p = strtok(NULL, ":"))
 	_add_load_path(p);
     free(tmp);
     return 0;
@@ -709,7 +808,7 @@ static int
 mod_init(int argc, char **argv)
 {
     SCM port;
-    
+
     scm_init_guile();
     scm_load_goops();
 
@@ -717,16 +816,18 @@ mod_init(int argc, char **argv)
 	return 1;
 
     _guile_init_strategy();
+    _guile_init_dico_key();
     _guile_init_dico_port();
     _guile_init_dico_log_port();
     _guile_init_funcs();
+#ifdef GUILE_DEBUG_MACROS
     if (guile_debug) {
 	SCM_DEVAL_P = 1;
 	SCM_BACKTRACE_P = 1;
 	SCM_RECORD_POSITIONS_P = 1;
 	SCM_RESET_DEBUG_MODE;
     }
-
+#endif
     port = _make_dico_log_port(L_ERR);
     if (port == SCM_BOOL_F) {
 	dico_log(L_ERR, 0, _("mod_init: cannot initialize error port"));
@@ -734,17 +835,17 @@ mod_init(int argc, char **argv)
     }
     scm_set_current_output_port(port);
     scm_set_current_error_port(port);
-    
+
     if (guile_init_script
 	&& guile_load(guile_init_script, guile_init_args)) {
-	dico_log(L_ERR, 0, _("mod_init: cannot load init script %s"), 
+	dico_log(L_ERR, 0, _("mod_init: cannot load init script %s"),
 		 guile_init_script);
 	return 1;
     }
 
     if (guile_init_fun && init_vtab(guile_init_fun, NULL, global_vtab))
 	return 1;
-    
+
     return 0;
 }
 
@@ -757,7 +858,7 @@ mod_init_db(const char *dbname, int argc, char **argv)
     char *init_script = NULL;
     char *init_args = NULL;
     char *init_fun = guile_init_fun;
-    
+
     struct dico_option db_option[] = {
 	{ DICO_OPTSTR(init-script), dico_opt_string, &init_script },
 	{ DICO_OPTSTR(init-args), dico_opt_string, &init_args },
@@ -769,13 +870,13 @@ mod_init_db(const char *dbname, int argc, char **argv)
 	return NULL;
     argc -= i;
     argv += i;
-    
+
     if (init_script && guile_load(init_script, init_args)) {
-	dico_log(L_ERR, 0, _("mod_init: cannot load init script %s"), 
+	dico_log(L_ERR, 0, _("mod_init: cannot load init script %s"),
 		 init_script);
 	return NULL;
     }
-    
+
     db = malloc(sizeof(*db));
     if (!db) {
 	memerr("mod_init_db");
@@ -805,12 +906,12 @@ mod_init_db(const char *dbname, int argc, char **argv)
 	    }
 	}
     }
-    
+
     if (err) {
 	free(db);
 	return NULL;
     }
-    
+
     db->argc = argc;
     db->argv = argv;
     return (dico_handle_t)db;
@@ -832,7 +933,7 @@ mod_close(dico_handle_t hp)
 
     if (db->vtab[close_proc])
 	if (guile_call_proc(&res, db->vtab[close_proc],
-			    scm_list_1(scm_cons(SCM_IM_QUOTE, db->handle))))
+			    scm_list_1(db->handle)))
 	    return 1;
     scm_gc_unprotect_object(db->handle);
 
@@ -845,9 +946,8 @@ argv_to_scm(int argc, char **argv)
     SCM scm_first = SCM_EOL, scm_last;
 
     for (; argc; argc--, argv++) {
-	SCM new = scm_cons(scm_cons(SCM_IM_QUOTE, scm_makfrom0str(*argv)), 
-                           SCM_EOL);
-	if (scm_first == SCM_EOL) 
+	SCM new = scm_cons(scm_from_locale_string(*argv), SCM_EOL);
+	if (scm_first == SCM_EOL)
 	    scm_last = scm_first = new;
 	else {
 	    SCM_SETCDR(scm_last, new);
@@ -863,14 +963,13 @@ assoc_to_scm(dico_assoc_list_t assoc)
     SCM scm_first = SCM_EOL, scm_last;
     dico_iterator_t itr;
     struct dico_assoc *p;
-	
+
     itr = dico_assoc_iterator(assoc);
     for (p = dico_iterator_first(itr); p; p = dico_iterator_next(itr)) {
-	SCM new = scm_cons(scm_cons(SCM_IM_QUOTE,
-				    scm_cons(scm_makfrom0str(p->key),
-					     scm_makfrom0str(p->value))), 
+	SCM new = scm_cons(scm_cons(scm_from_locale_string(p->key),
+				    scm_from_locale_string(p->value)),
 			   SCM_EOL);
-	if (scm_first == SCM_EOL) 
+	if (scm_first == SCM_EOL)
 	    scm_last = scm_first = new;
 	else {
 	    SCM_SETCDR(scm_last, new);
@@ -903,9 +1002,8 @@ mod_open(dico_handle_t dp)
 {
     struct _guile_database *db = (struct _guile_database *)dp;
     if (guile_call_proc(&db->handle, db->vtab[open_proc],
-			scm_cons(scm_cons(SCM_IM_QUOTE,
-					  scm_makfrom0str(db->dbname)),
-				 argv_to_scm(db->argc, db->argv)))) 
+			scm_cons(scm_from_locale_string(db->dbname),
+				 argv_to_scm(db->argc, db->argv))))
 	return 1;
     if (db->handle == SCM_EOL || db->handle == SCM_BOOL_F)
 	return 1;
@@ -918,11 +1016,10 @@ mod_get_text(struct _guile_database *db, int n)
 {
     if (db->vtab[n]) {
 	SCM res;
-	
-	if (guile_call_proc(&res, db->vtab[n],
-			    scm_list_1(scm_cons(SCM_IM_QUOTE, db->handle))))
+
+	if (guile_call_proc(&res, db->vtab[n], scm_list_1(db->handle)))
 	    return NULL;
-	if (scm_is_string(res)) 
+	if (scm_is_string(res))
 	    return scm_to_locale_string(res);
 	else {
 	    rettype_error(db->vtab[n]);
@@ -930,7 +1027,7 @@ mod_get_text(struct _guile_database *db, int n)
 	}
     }
     return NULL;
-}    
+}
 
 static char *
 mod_info(dico_handle_t hp)
@@ -960,7 +1057,7 @@ scm_to_langlist(SCM scm, SCM procsym)
 	list = dico_list_create();
 	for (; scm != SCM_EOL && scm_is_pair(scm); scm = SCM_CDR(scm))
 	    dico_list_append(list, scm_to_locale_string(SCM_CAR(scm)));
-    } else 
+    } else
 	rettype_error(procsym);
     return list;
 }
@@ -973,9 +1070,8 @@ mod_lang(dico_handle_t hp, dico_list_t list[2])
     list[0] = list[1] = NULL;
     if (proc) {
 	SCM res;
-	
-	if (guile_call_proc(&res, proc,
-			    scm_list_1(scm_cons(SCM_IM_QUOTE, db->handle))))
+
+	if (guile_call_proc(&res, proc, scm_list_1(db->handle)))
 	    return 1;
 	if (res == SCM_EOL)
 	    /* ok, nothing */;
@@ -1018,22 +1114,21 @@ mod_match(dico_handle_t hp, const dico_strategy_t strat, const char *word)
     struct _guile_database *db = (struct _guile_database *)hp;
     SCM scm_strat = _make_strategy(strat);
     SCM res;
+    struct dico_key *key;
+    SCM scm_key;
 
-    if (strat->sel
-	&& strat->sel(DICO_SELECT_BEGIN, word, NULL, strat->closure)) {
-	dico_log(L_ERR, 0, _("mod_match: initial select failed"));
+    scm_key = dico_new_scm_key(&key);
+    
+    if (dico_key_init(key, strat, word)) {
+	dico_log(L_ERR, 0, _("mod_match: key initialization failed"));
 	return NULL;
     }
-    
+
     if (guile_call_proc(&res, db->vtab[match_proc],
-			scm_list_3(scm_cons(SCM_IM_QUOTE, db->handle),
-				   scm_cons(SCM_IM_QUOTE, scm_strat),
-				   scm_cons(SCM_IM_QUOTE,
-					    scm_makfrom0str(word)))))
+			scm_list_3(db->handle, scm_strat, scm_key)))
 	return NULL;
 
-    if (strat->sel)
-	strat->sel(DICO_SELECT_END, word, NULL, strat->closure);
+    dico_key_deinit(key);
 
     if (res == SCM_BOOL_F || res == SCM_EOL)
 	return NULL;
@@ -1046,11 +1141,10 @@ mod_define(dico_handle_t hp, const char *word)
 {
     struct _guile_database *db = (struct _guile_database *)hp;
     SCM res;
-	
+
     if (guile_call_proc(&res, db->vtab[define_proc],
-			scm_list_2(scm_cons(SCM_IM_QUOTE, db->handle),
-				   scm_cons(SCM_IM_QUOTE,
-					    scm_makfrom0str(word)))))
+			scm_list_2(db->handle,
+				   scm_from_locale_string(word))))
 	return NULL;
 
     if (res == SCM_BOOL_F || res == SCM_EOL)
@@ -1067,13 +1161,11 @@ mod_output_result (dico_result_t rp, size_t n, dico_stream_t str)
     SCM res;
     SCM oport = scm_current_output_port();
     SCM port = _make_dico_port(str);
-    
+
     scm_set_current_output_port(port);
-    
+
     rc = guile_call_proc(&res, gres->db->vtab[output_proc],
-			 scm_list_2(scm_cons(SCM_IM_QUOTE, gres->result),
-				    scm_cons(SCM_IM_QUOTE,
-					     scm_from_int(n))));
+			 scm_list_2(gres->result, scm_from_int(n)));
     scm_set_current_output_port(oport);
     _dico_port_close(port);
     if (rc)
@@ -1086,11 +1178,11 @@ mod_result_count (dico_result_t rp)
 {
     struct guile_result *gres = (struct guile_result *)rp;
     SCM res;
-    
+
     if (guile_call_proc(&res, gres->db->vtab[result_count_proc],
-			scm_list_1(scm_cons(SCM_IM_QUOTE, gres->result))))
+			scm_list_1(gres->result)))
 	return 0;
-    if (scm_is_integer(res)) 
+    if (scm_is_integer(res))
 	return scm_to_int32(res);
     else
 	rettype_error(gres->db->vtab[result_count_proc]);
@@ -1104,13 +1196,13 @@ mod_compare_count (dico_result_t rp)
 
     if (gres->db->vtab[compare_count_proc]) {
 	SCM res;
-	
+
 	if (guile_call_proc(&res, gres->db->vtab[compare_count_proc],
-			    scm_list_1(scm_cons(SCM_IM_QUOTE, gres->result))))
+			    scm_list_1(gres->result)))
 	    return 0;
-	if (scm_is_integer(res)) 
+	if (scm_is_integer(res))
 	    return scm_to_int32(res);
-	else 
+	else
 	    rettype_error(gres->db->vtab[compare_count_proc]);
     }
     return 0;
@@ -1123,9 +1215,9 @@ mod_free_result(dico_result_t rp)
 
     if (gres->db->vtab[free_result_proc]) {
 	SCM res;
-	
+
 	guile_call_proc(&res, gres->db->vtab[free_result_proc],
-			scm_list_1(scm_cons(SCM_IM_QUOTE, gres->result)));
+			scm_list_1(gres->result));
     }
     scm_gc_unprotect_object(gres->result);
     free(gres);
@@ -1136,13 +1228,12 @@ mod_result_headers (dico_result_t rp, dico_assoc_list_t hdr)
 {
     struct guile_result *gres = (struct guile_result *)rp;
     SCM proc = gres->db->vtab[result_headers_proc];
-    
+
     if (proc) {
 	SCM res;
-	
+
 	if (guile_call_proc(&res, proc,
-			    scm_list_2(scm_cons(SCM_IM_QUOTE, gres->result),
-				       scm_cons(SCM_IM_QUOTE,assoc_to_scm(hdr)))))
+			    scm_list_2(gres->result, assoc_to_scm(hdr))))
 	    return 1;
 	if (!scm_is_pair(res)) {
 	    rettype_error(proc);
