@@ -1,5 +1,5 @@
 /* This file is part of GNU Dico.
-   Copyright (C) 2008 Sergey Poznyakoff
+   Copyright (C) 2008, 2010 Sergey Poznyakoff
 
    GNU Dico is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -262,7 +262,7 @@ revert_word(char *dst, const char *src, size_t len)
     char *p = dst + len;
 
     *p = 0;
-    for (utf8_iter_first(&itr, (unsigned char *)src);
+    for (utf8_iter_first(&itr, (char *)src);
 	 !utf8_iter_end_p(&itr);
 	 utf8_iter_next(&itr)) {
 	p -= itr.curwidth;
@@ -654,14 +654,14 @@ outline_match0(dico_handle_t hp, const char *strat,
 }
 
 dico_result_t
-outline_match_all(dico_handle_t hp, const char *word,
-		  dico_select_t sel, void *closure)
+outline_match_all(dico_handle_t hp, dico_strategy_t strat, const char *word)
 {
     struct outline_file *file = (struct outline_file *) hp;
     dico_list_t list;
     size_t count, i;
     struct result *res;
-    
+    struct dico_key key;
+
     list = dico_list_create();
 
     if (!list) {
@@ -669,18 +669,19 @@ outline_match_all(dico_handle_t hp, const char *word,
 	return NULL;
     }
 
-    if (sel(DICO_SELECT_BEGIN, word, NULL, closure)) {
-	dico_log(L_ERR, 0, _("outline_match_all: initial select failed"));
+    if (dico_key_init(&key, strat, word)) {
+	dico_log(L_ERR, 0, _("outline_match_all: key initialization failed"));
 	return NULL;
     }
     
     for (i = 0; i < file->count; i++) {
-	if (sel(DICO_SELECT_RUN, word, file->index[i].word, closure)) {
+	if (dico_key_match(&key, file->index[i].word))
 	    dico_list_append(list, &file->index[i]);
-	}
     }
+
+    dico_key_deinit(&key);
+    
     compare_count = file->count;
-    sel(DICO_SELECT_END, word, NULL, closure);
 	
     count = dico_list_count(list);
     if (count == 0) {
@@ -703,7 +704,7 @@ dico_result_t
 outline_match(dico_handle_t hp, const dico_strategy_t strat, const char *word)
 {
     if (strat->sel) 
-	return outline_match_all(hp, word, strat->sel, strat->closure);
+	return outline_match_all(hp, strat, word);
     else
 	return outline_match0(hp, strat->name, word);
 }
