@@ -1,5 +1,5 @@
 ;;;; This file is part of GNU Dico.
-;;;; Copyright (C) 2008 Sergey Poznyakoff
+;;;; Copyright (C) 2008, 2010 Sergey Poznyakoff
 ;;;; 
 ;;;; GNU Dico is free software; you can redistribute it and/or modify
 ;;;; it under the terms of the GNU General Public License as published by
@@ -35,7 +35,7 @@
 (define (define-word dbh word)
   #f)
 
-(define (match-word dbh strat word)
+(define (match-word dbh strat key)
   #f)
 
 (define (output rh n)
